# (c) 2005-2020 Ali Afshar <aafshar@gmail.com>.
# MIT License. See LICENSE.
# vim: ft=python sw=2 ts=2 sts=2 tw=80

"""Simple logging to stdout/stderr."""

import datetime, pprint

class Level:
  DEBUG = 0
  INFO = 1
  ERROR = 2
  CRITICAL = 3

  @staticmethod
  def name(level: int) -> str:
    return {
        Level.DEBUG: 'D',
        Level.INFO: 'I',
        Level.ERROR: 'E',
        Level.CRITICAL: 'C',
    }.get(level)

    


class LoggerMixin:

  level = Level.INFO

  @classmethod
  def set_level(cls, lstr):
    LoggerMixin.level = {
        'debug': Level.DEBUG,
        'info': Level.INFO,
        'error': Level.ERROR,
        'critical': Level.CRITICAL,
    }.get(lstr)

  def __init__(self):
    self.name = self.__gtype_name__

  def msg(self, msg: str, level: int):
    if level >= LoggerMixin.level:
      print(f'{Level.name(level)}:{self.name}:{msg}')

  def debug(self, msg: str, data=None):
    if LoggerMixin.level:
      return
    if data:
      msg = f'{msg} ->\n{pprint.pformat(data)}\n<-'
    self.msg(msg, Level.DEBUG)

  def info(self, msg: str):
    self.msg(msg, Level.INFO)

  def error(self, msg: str):
    self.msg(msg, Level.ERROR)




# (c) 2005-2020 Ali Afshar <aafshar@gmail.com>.
# MIT License. See LICENSE.
# vim: ft=python sw=2 ts=2 sts=2 tw=80

"""The Bominade basic widgets."""

import os

from typing import Iterable

from gi.repository import Gio, GLib, GObject, Gdk, Gtk



class Base(Gtk.EventBox):
  """Base B8 Widget."""

  def __init__(self, b8):
    self.b8 = b8
    Gtk.EventBox.__init__(self)

  def create_ui(self):
    """Override to create the UI."""
    raise NotImplementedError


class BaseButtonMixin:
  """Mixin for button operations."""

  def create_button(self, key: str, icon: str, tooltip: str,
                    size: Gtk.IconSize=Gtk.IconSize.SMALL_TOOLBAR):
    """Creating the content for buttons is always the same."""
    self.key = key
    im = Gtk.Image.new_from_icon_name(icon, size)
    self.set_image(im)
    self.set_tooltip_text(tooltip)


class ImageButton(Gtk.Button, BaseButtonMixin):
  """Button with an image."""

  __gtype_name__ = 'b8-imagebutton'

  key = GObject.Property(type=str)

  def __init__(self, *args, **kw):
    Gtk.Button.__init__(self)
    self.create_button(*args, **kw)


class ImageToggleButton(Gtk.ToggleButton, BaseButtonMixin):
  """Toggle button with an image."""

  __gtype_name__ = 'b8-imagetogglebutton'

  name = GObject.Property(type=str)

  def __init__(self, *args, **kw):
    Gtk.ToggleButton.__init__(self)
    self.create_button(*args, **kw)


class MiniToolbar(Gtk.Box):

  __gtype_name__ = 'b8-minitoolbar'

  __gsignals__ = {
    'clicked': (GObject.SignalFlags.RUN_FIRST, None, (Gtk.Button, str,)),
    'right-clicked': (GObject.SignalFlags.RUN_FIRST, None, (Gtk.Button, str,))
  }

  def __init__(self, buttons: Iterable[Gtk.Widget], orientation: Gtk.Orientation, spacing: int=0):
    Gtk.Box.__init__(self, orientation=orientation, spacing=spacing)
    for b in buttons:
      if hasattr(b, 'key'):
        b.connect('clicked', self.on_clicked)
        b.connect('button-press-event', self.on_button_press_event)
        self.pack_start(b, False, False, 0)
      else:
        self.pack_start(b, True, True, 2)

  @staticmethod
  def spacer():
    return Gtk.Frame()

  @classmethod
  def horizontal(cls, buttons):
    return cls(buttons, Gtk.Orientation.HORIZONTAL)

  @classmethod
  def vertical(cls, buttons):
    return cls(buttons, Gtk.Orientation.VERTICAL)

  def on_clicked(self, w):
    self.emit('clicked', w, w.key)

  def on_button_press_event(self, w, event):
    if event.button == Gdk.BUTTON_SECONDARY:
      self.emit('right-clicked', w, w.key)


class PopupMenuItem(Gtk.MenuItem):
  """Menu item for a popup menu."""

  __gtype_name__ = 'b8-popupmenuitem'

  def __init__(self, key, icon, text, size=Gtk.IconSize.SMALL_TOOLBAR):
    self.key = key
    Gtk.MenuItem.__init__(self)
    b = Gtk.HBox()
    self.add(b)
    i = Gtk.Image.new_from_icon_name(icon, size)
    b.pack_start(i, False, False, 0)
    l = Gtk.Label()
    l.set_label(text)
    b.pack_start(l, False, False, 12)


class FilesystemPopupMenu(Gtk.Menu):

  SEPARATOR = object()

  __gtype_name__ = 'b8-fspopupmenu'

  __gsignals__ = {
    'activate': (GObject.SignalFlags.RUN_FIRST, None, (Gtk.MenuItem, str, Gio.File)),
  }

  gfile = GObject.Property(type=Gio.File)

  def __init__(self, gfile, suppress=None):
    Gtk.Menu.__init__(self)
    self.gfile = gfile
    title = PopupMenuItem(
        key='title',
        icon='open-menu',
        text=gfile.get_basename(),
    )
    title.set_sensitive(False)
    self.add(title)
    self.add(FilesystemPopupMenu.separator())
    if not suppress:
      suppress = set()
    for i in self.items():
      self.add(i)
      if hasattr(i, 'key'):
        i.connect('activate', self.on_activate)
        if i.key in suppress:
          i.set_sensitive(False)

  def items(self) -> Iterable[Gtk.MenuItem]:
    """Override to return menu items."""
    raise NotImplementedError()

  def popup(self, event):
    self.show_all()
    Gtk.Menu.popup(self, None, None, None, None, event.button, event.time)

  def on_activate(self, w):
    self.emit('activate', w, w.key, self.gfile)

  @staticmethod
  def separator():
    return Gtk.SeparatorMenuItem()


class FilePopupMenu(FilesystemPopupMenu):
  """Popup menu for files."""

  __gtype_name__ = 'b8-filepopupmenu'

  def items(self) -> Iterable[Gtk.MenuItem]:
    return [
      PopupMenuItem(
        key='open',
        icon='document-open',
        text='Open file',
      ),
      FilesystemPopupMenu.separator(),
      PopupMenuItem(
        key='browseparent',
        icon='folder-open',
        text='Browse parent directory',
      ),
      PopupMenuItem(
        key='terminalparent',
        icon='utilities-terminal',
        text='Terminal in parent directory',
      ),
      FilesystemPopupMenu.separator(),
      PopupMenuItem(
        key='close',
        icon='window-close',
        text='Close file',
      ),
    ]


class DirectoryPopupMenu(FilesystemPopupMenu):

  __gtype_name__ = 'b8-directorypopupmenu'

  def items(self):
    return [
      PopupMenuItem(
        key='browse',
        icon='folder-open',
        text='Browse directory',
      ),
      PopupMenuItem(
        key='terminal',
        icon='utilities-terminal',
        text='Terminal in directory',
      ),
    ]


class MenuHandlerMixin:

  def _on_menu_activate(self, w, m, key, gfile):
    keyhandlers = {
        'open': self._on_open_activate,
        'browseparent': self._on_browseparent_activate,
        'terminalparent': self._on_terminalparent_activate,
        'close': self._on_close_activate,
        'browse': self._on_browse_activate,
        'terminal': self._on_terminal_activate,
    }
    f = keyhandlers.get(key)
    if f:
      f(m, gfile)

  def _on_open_activate(self, w, gfile):
    self.emit('file-activated', gfile)

  def _on_browseparent_activate(self, w, gfile):
    self.emit('directory-activated', gfile.get_parent())

  def _on_terminalparent_activate(self, w, gfile):
    self.emit('terminal-activated', gfile.get_parent())

  def _on_close_activate(self, w, gfile):
    self.emit('file-destroyed', gfile)

  def _on_browse_activate(self, w, gfile):
    self.emit('directory-activated', gfile)

  def _on_terminal_activate(self, w, gfile):
    self.emit('terminal-activated', gfile)



def parse_color(s: str) -> Gdk.RGBA:
  """Parse a color string."""
  c = Gdk.RGBA()
  c.parse(s)
  return c


class Colors:
  """The default colors used in B8."""

  RED = parse_color('#d30102')
  ORANGE = parse_color('#cb4b16')
  GREEN = parse_color('#859900')





# (c) 2005-2020 Ali Afshar <aafshar@gmail.com>.
# MIT License. See LICENSE.
# vim: ft=python sw=2 ts=2 sts=2 tw=80


"""The Bominade file browser."""


import os

from gi.repository import Gio, GLib, GObject, Gdk, Gtk, GdkPixbuf, Pango

from b8 import ui, vim, logs



class Buffers(Gtk.ScrolledWindow, ui.MenuHandlerMixin, logs.LoggerMixin):

  __gtype_name__ = 'b8-buffers'

  __gsignals__ = {
    'directory-changed': (GObject.SignalFlags.RUN_FIRST, None, (Gio.File,)),
    'file-activated': (GObject.SignalFlags.RUN_FIRST, None, (Gio.File,)),
    'file-destroyed': (GObject.SignalFlags.RUN_FIRST, None, (Gio.File,)),
    'terminal-activated': (GObject.SignalFlags.RUN_FIRST, None, (Gio.File,)),
    'directory-activated': (GObject.SignalFlags.RUN_FIRST, None, (Gio.File,)),
    'buffer-activated':(GObject.SignalFlags.RUN_FIRST, None, (vim.Buffer,)),
  }
 
  
  def __init__(self):
    Gtk.ScrolledWindow.__init__(self)
    logs.LoggerMixin.__init__(self)
    self.model = self._create_model()
    self.tree = self._create_tree(self.model)
    self.add(self.tree)

  def _create_model(self):
    m = Gtk.ListStore(object, str)
    return m

  def _create_tree(self, model):
    t = Gtk.TreeView(model=model)
    t.set_headers_visible(False)
    ce = Gtk.CellRendererText()
    ce.set_padding(6, 1)
    co = Gtk.TreeViewColumn('name', ce)
    co.add_attribute(ce, 'markup', 1)
    t.append_column(co)
    t.connect('row-activated', self.on_row_activated)
    t.connect('button-press-event', self.on_button_press_event)
    t.set_activate_on_single_click(True)
    return t

  def next(self):
    l = self.model.iter_n_children(None)
    if not l:
      return
    m, siter = self.tree.get_selection().get_selected()
    giter = self.model.iter_next(siter)
    if not giter:
      # We are at the end or there is only 1, doesn't matter
      giter = self.model.iter_nth_child(None, 0)
    self.activate_iter(giter)

  def prev(self):
    l = self.model.iter_n_children(None)
    if not l:
      return
    m, siter = self.tree.get_selection().get_selected()
    giter = self.model.iter_previous(siter)
    if not giter:
      # We are at the beginning or there is only 1, doesn't matter
      giter = self.model.iter_nth_child(None, l - 1)
    self.activate_iter(giter)

  def activate_iter(self, giter):
    b = self.model.get_value(giter, 0)
    self.emit('buffer-activated', b)
    
  def select(self, giter):
    selection = self.tree.get_selection()
    m, siter = selection.get_selected()
    if siter != giter:
      self.tree.get_selection().select_iter(giter)
    b = self.model.get(giter, 0)[0]
    self.debug(f'select {b}')
    #self.b8.ui.window.set_title(b.path)
    #self.b8.ui.vimview.drawing_area.grab_focus()

  def change(self, f, number):
    self.debug(f'buffer change {f.get_path()} {number}')
    for grow in self.model:
      b = self.model.get_value(grow.iter, 0)

      if b.number == number and b.path == f.get_path():
        self.debug(f'existing {b}')
        self.select(grow.iter)
        return
    b = vim.Buffer(number, f)
    giter = self.model.append([b, b.markup])
    self.select(giter)
    #self.b8.sessions.buffers.append(b.path)
    #self.b8.sessions.save()

  def remove(self, f, number):
    for grow in self.model:
      b = self.model.get_value(grow.iter, 0)
      print(b.path, f.get_path())
      if b.number == number and b.path == os.path.realpath(f.get_path()):
        self.model.remove(grow.iter)
        #self.b8.sessions.buffers.remove(b.path)
        #self.b8.sessions.save()
        return


  def on_row_activated(self, w, path, column):
    giter = self.model.get_iter(path)
    b = self.model.get(giter, 0)[0]
    #self.b8.vim.nvim_change_buffer(b.number)
    self.emit('buffer-activated', b)
    self.debug(f'row-activated {b}')

  def render(self, cell_layout, cell, tree_model, iter, *data):
    b = tree_model.get_value(iter, 0)
    cell.set_property('markup', b.markup)

  def on_button_press_event(self, treeview, event):
    if event.button != Gdk.BUTTON_SECONDARY:
      return
    item_spec = self.tree.get_path_at_pos(int(event.x), int(event.y))
    if item_spec is not None:
      # clicked on an actual cell
      path, col, rx, ry = item_spec
      giter = self.model.get_iter(path)
      b = self.model.get_value(giter, 0)
      menu = ui.FilePopupMenu(b.file)
      menu.connect('activate', self._on_menu_activate)
      menu.popup(event)
      return True

if __name__ == '__main__':
  w = Gtk.Window()
  w.connect('destroy', Gtk.main_quit)
  fs = Buffers()
  w.add(fs)
  w.resize(400, 400)
  w.show_all()

  fs.change(Gio.File.new_for_path('/home/aa/.vimrc'), 1)
  fs.change(Gio.File.new_for_path('/home/aa/.zshrc'), 1)


  def term_activated(w, f):
    print('term-activated', w, f, f.get_path())

  def file_activated(w, f):
    print('file-activated', w, f, f.get_path())

  def directory_changed(w, f):
    print('directory-changed', w, f, f.get_path())

  def directory_activated(w, f):
    print('directory activate')
    w.browse(f)

  def buffer_activated(w, b):
    print('buffer-activated', b)

  def file_destroyed(w, f):
    print('file-destroyed', w, f, f.get_path())

  fs.connect('terminal-activated', term_activated)
  fs.connect('file-activated', file_activated)
  fs.connect('directory-changed', directory_changed)
  fs.connect('directory-activated', directory_activated)
  fs.connect('file-destroyed', file_destroyed)
  fs.connect('buffer-activated', buffer_activated)


  Gtk.main()



# (c) 2005-2020 Ali Afshar <aafshar@gmail.com>.
# MIT License. See LICENSE.
# vim: ft=python sw=2 ts=2 sts=2 tw=80


from gi.repository import Gio, GLib, GObject

import os, argparse, configparser

import b8
from b8 import logs



class Item(GObject.GObject):

  __gtype_name__ = 'b8-config-item'

  section = GObject.Property(type=str)
  name = GObject.Property(type=str)
  default = GObject.Property(type=str)
  value = GObject.Property(type=str)
  description = GObject.Property(type=str)
  short = GObject.Property(type=str)

  def __init__(self, section: str, name: str, default: str, description: str):
    GObject.GObject.__init__(self)
    self.section = section
    self.name = name
    self.default = default
    self.description = description

  def prime_argparser(self, p):
    p.add_argument(self.arg_name, help=self.description)

  def read_argparser(self, ns):
    return getattr(ns, self.ns_key)

  def read_configparser(self, p):
    return p.get(self.section, self.name, fallback=None)

  def read_env(self, e):
    pass

  @property
  def key(self):
    return (self.section, self.name)

  @property
  def arg_name(self):
    return f'--{self.section}-{self.name}'

  @property
  def ns_name(self):
    return self.name.replace('-', '_')

  @property
  def ns_key(self):
    return f'{self.section}_{self.ns_name}'




class Instance(GObject.GObject):

  __gtype_name__  = 'b8-config-instance'

  def __init__(self):
    GObject.GObect.__init__(self)

  def init(self):
    self.root_path = os.path.expanduser('~/.config/b8')
    self.run_path = os.path.join(self.root_path, 'run')
    self.config_path = os.path.join(self.root_path, 'b8rc')
    self.sessions_path = os.path.join(self.root_path, 'session.json')
    self.create()
  
  # Switch to Gio
  def create(self):
    try:
      os.makedirs(self.root_path)
      self.debug('create config directory {self.root_path}')
    except FileExistsError:
      self.debug(f'config directory exists {self.root_path}')
    try:
      os.makedirs(self.run_path)
      self.debug('create run directory {self.run_path}')
    except FileExistsError:
      self.debug(f'run exists {self.root_path}')
    if not os.path.exists(self.config_path):
      with open(self.config_path, 'w') as f:
        f.write('# Bominade Config File\n')


CONFIG_DIR = Gio.File.new_for_path(GLib.get_user_config_dir()).get_child('b8')
CONFIG_FILE = CONFIG_DIR.get_child('b8.ini')
CONFIG_EMPTY = '# Bominade Config File\n##\n'.encode('utf-8')

class ConfigError(RuntimeError):
  """Error with configuration."""


class Config(GObject.GObject, logs.LoggerMixin):

  __gtype_name__ = 'b8-config'

  root = GObject.Property(type=Gio.File, default=CONFIG_DIR)
  file = GObject.Property(type=Gio.File, default=CONFIG_FILE)

  items = [
      Item('logging', 'level', 'info',
        'The logging level to use'),
      Item('terminal', 'theme', 'b8',
        'The terminal theme to use'),
      Item('terminal', 'font', 'Monospace 13',
        'The terminal font to use, e.g. "Monospace 13"'),
      Item('shortcuts', 'previous-buffer', '<Alt>Up',
        'Shortcut key to switch to the previous buffer'),
      Item('shortcuts', 'next-buffer', '<Alt>Down',
        'Shortcut key to switch to the next buffer'),
      Item('shortcuts', 'previous-terminal', '<Alt>Left',
        'Shortcut key to switch to the previous terminal'),
      Item('shortcuts', 'next-terminal', '<Alt>Right',
        'Shortcut key to switch to the next terminal'),
      Item('shortcuts', 'new-terminal', '<Alt>t', 
        'Shortcut key to create a new terminal'),
  ]

  def __init__(self):
    GObject.GObject.__init__(self)
    logs.LoggerMixin.__init__(self)
    self._create_root()
    self.values = {}
    self._read_default()
    ns = self._read_args()
    self._read_configparser()
    self._read_argparser(ns)
    logs.LoggerMixin.set_level(self.values['logging', 'level'])
    self.debug(f'root user directory is at {self.root.get_path()}')
    self.debug('values', data=self.values)

  def _create_root(self):
    try:
      self.root.make_directory_with_parents(None)
    except GLib.Error as e:
      if e.code != Gio.IOErrorEnum.EXISTS:
        raise
    if not self.file.query_exists(None):
      f = self.file.create_readwrite(Gio.FileCreateFlags.NONE, None)
      s = f.get_output_stream()
      s.write(CONFIG_EMPTY)
      s.close()

  def _read_default(self):
    for item in self.items:
      self.values[item.key] = item.default

  def _read_configparser(self):
    p = configparser.ConfigParser()
    p.read(self.file.get_path())
    for item in self.items:
      v = item.read_configparser(p)
      if v is not None:
        self.values[item.key] = v

  def _read_args(self):
    p = argparse.ArgumentParser(
        prog = 'b8',
        description = f'The bominade IDE, version {b8.__version__}',
    )
    p.add_argument('-d', '--debug', action='store_true',
        help='Run with logging level debug')
    p.add_argument('-f', '--config', help='Configuration file to use',
        default=self.file.get_path())
    for item in self.items:
      item.prime_argparser(p)
    ns = p.parse_args()
    if ns.config:
      self.file = Gio.File.new_for_path(os.path.expanduser(ns.config))
      if not self.file.query_exists(None):
        e = f'config file "{ns.config}" does not exist'
        self.error(e)
        raise ConfigError(e)
    if ns.debug:
      ns.logging_level = 'debug'
    return ns

  def _read_argparser(self, ns):
    for item in self.items:
      v = item.read_argparser(ns)
      if v is not None:
        self.values[item.key] = v

  def get(self, key):
    return self.values.get(key)


if __name__ == '__main__':
  c = Config()



# (c) 2005-2020 Ali Afshar <aafshar@gmail.com>.
# MIT License. See LICENSE.
# vim: ft=python sw=2 ts=2 sts=2 tw=80

"""The Bominade file browser."""

import os

from gi.repository import Gio, GLib, GObject, Gdk, Gtk, GdkPixbuf, Pango

from b8 import ui


DEFAULT_FILE_ATTRIBUTES = '*'


class FileListItem:
  """Item to go in the file list.

  This is a thin wrapper over a Gio.FileInfo.
  """

  def __init__(self, info: Gio.FileInfo, parent: Gio.File):
    self.info = info
    self.parent = parent
    self.name = info.get_name()
    self.path = parent.get_child(self.name).get_path()
    self.file = parent.get_child(self.name)
    icon_to_load = Gtk.IconTheme.get_default().choose_icon(
        self.info.get_icon().get_names(),
        Gtk.IconSize.SMALL_TOOLBAR, 0
    )
    if icon_to_load:
      self.icon = icon_to_load.load_icon()
    else:
      self.icon = None
    self.file_type = self.info.get_file_type()
    self.is_directory = self.file_type == Gio.FileType.DIRECTORY
    if self.is_directory:
      self.sort_prefix = 0
    else:
      self.sort_prefix = 1
    self.sort_key = f'{self.sort_prefix}_{self.name}'


class Files(Gtk.VBox, ui.MenuHandlerMixin):
  """File browser widget."""

  __gtype_name__ = 'b8-files'

  __gsignals__ = {
    'directory-changed': (GObject.SignalFlags.RUN_FIRST, None, (Gio.File,)),
    'file-activated': (GObject.SignalFlags.RUN_FIRST, None, (Gio.File,)),
    'file-destroyed': (GObject.SignalFlags.RUN_FIRST, None, (Gio.File,)),
    'terminal-activated': (GObject.SignalFlags.RUN_FIRST, None, (Gio.File,)),
    'directory-activated': (GObject.SignalFlags.RUN_FIRST, None, (Gio.File,)),
  }

  directory = GObject.Property(type=Gio.File)
  show_hidden = GObject.Property(type=bool, default=False)

  def __init__(self):
    """Generate the user interface."""
    Gtk.VBox.__init__(self)
    self.title = self.create_title_label()
    self.pack_start(self.create_toolbar(), False, False, 0)
    c = Gtk.ScrolledWindow()
    self.pack_start(c, True, True, 0)
    self.model = self.create_model()
    self.tree = self.create_tree(self.model)
    c.add(self.tree)
    self.connect('notify::directory', self.on_directory_notify)

  def create_model(self):
    """Create the tree model."""
    m = Gtk.ListStore(
        object,
        GObject.TYPE_STRING, # sort key
        GdkPixbuf.Pixbuf, # icon
        GObject.TYPE_STRING, # name
        GObject.TYPE_STRING, # mod
        Gdk.RGBA, # mod color
    )
    m.set_sort_column_id(1, Gtk.SortType.ASCENDING)
    return m

  def create_tree(self, m: Gtk.ListStore):
    """Create the tree view."""
    t = Gtk.TreeView(m)
    t.set_activate_on_single_click(False)
    t.set_headers_visible(False)
    t.connect('row-activated', self.on_row_activated)
    t.connect('button-press-event', self.on_button_press_event)

    ce_icon = Gtk.CellRendererPixbuf()
    co_icon = Gtk.TreeViewColumn('Icon', ce_icon)
    co_icon.add_attribute(ce_icon, 'pixbuf', 2)
    t.append_column(co_icon)

    ce_mods = Gtk.CellRendererText()
    ce_mods.set_property('weight', 800)
    co_mods = Gtk.TreeViewColumn('Modifiers', ce_mods)
    co_mods.add_attribute(ce_mods, 'markup', 4)
    co_mods.add_attribute(ce_mods, 'foreground-rgba', 5)
    t.append_column(co_mods)

    ce_name = Gtk.CellRendererText()
    co_name = Gtk.TreeViewColumn('Filename', ce_name)
    co_name.set_expand(True)
    co_name.add_attribute(ce_name, 'text', 3)
    t.append_column(co_name)
    return t

  def create_title_label(self):
    """Create the current directory label"""
    l = Gtk.Label()
    l.set_ellipsize(Pango.EllipsizeMode.START)
    return l

  def create_toolbar(self):
    """Create the mini toolbar."""
    self.title_label = Gtk.Label()
    t = ui.MiniToolbar.horizontal(
        [
          ui.ImageButton(
            key='up',
            icon='go-up',
            tooltip='Navigate to the parent directory',
          ),
          ui.ImageButton(
            key='refresh',
            icon='view-refresh',
            tooltip='Refresh the current directory',
          ),
          ui.ImageButton(
            key='terminal',
            icon='utilities-terminal',
            tooltip='Start a terminal in this directory',
          ),
          self.title,
          ui.ImageToggleButton(
            key='hidden',
            icon='view-more',
            tooltip='Show or hide hidden files',
          ),
        ]
    )
    t.connect('clicked', self.on_toolbar_clicked)
    return t

  def browse_path(self, path: str, refresh: bool=False):
    """Browse the given path."""
    if self.directory and self.directory.get_path() == path:
      if not refresh:
        return

    f = Gio.File.new_for_path(path)
    f.enumerate_children_async(
      DEFAULT_FILE_ATTRIBUTES,
      Gio.FileQueryInfoFlags.NONE,
      GLib.PRIORITY_DEFAULT,
      None, # cancellable
      self.browse_path_callback,
      f,
    )

  def browse(self, gfile: Gio.File, refresh: bool=True):
    self.browse_path(gfile.get_path(), refresh=refresh)

  def refresh(self):
    """Refresh the current path."""
    self.browse(self.directory, refresh=True)

  def browse_path_callback(self, src, res, parent: Gio.File) -> None:
    """Async callback for finishing a File list enumeration."""
    self.model.clear()
    self.directory = parent
    for file_info in src.enumerate_children_finish(res):
      if not self.show_hidden and file_info.get_is_hidden():
        continue
      f = FileListItem(file_info, parent)
      self.append(f)
    self.git_revparse()

  def append(self, f):
    """Append an item to the model."""
    self.model.append([f, f.sort_key, f.icon, f.name, '', None])

  def git_revparse_callback(self, src, res):
    """Async callback for revparse command."""
    success, stdout, stderr = src.communicate_finish(res)
    if not success:
      return
    d = stdout.get_data().strip().decode('utf-8')
    self.git_status(Gio.File.new_for_path(d))

  def git_revparse(self):
    """Get the Git root directory."""
    l = Gio.SubprocessLauncher()
    l.set_flags(Gio.SubprocessFlags.STDOUT_PIPE | Gio.SubprocessFlags.STDERR_PIPE)
    l.set_cwd(self.directory.get_path())
    p = l.spawnv(['git', 'rev-parse', '--show-toplevel'])
    p.communicate_async(None, None, self.git_revparse_callback)

  def git_status_callback(self, src, res, parent):
    """Async callback for finishing a git status call."""
    success, stdout, stderr = src.communicate_finish(res)
    if not success:
      return
    d = stdout.get_data().decode('utf-8').splitlines()
    mods = {}
    cols = {
      'M': ui.Colors.RED,
      '??': ui.Colors.GREEN,
    }
    for line in d:
      line = line.strip().rstrip('/')
      mod, name = line.split()
      path = parent.get_child(name).get_path()
      mods[path] = mod
    for row in self.model:
      path = self.model.get_value(row.iter, 0).file.get_path()
      mod = mods.get(os.path.realpath(path))
      if mod:
        self.model.set_value(row.iter, 4, mod)
        c = cols.get(mod, ui.Colors.GREEN)
        self.model.set_value(row.iter, 5, c)

  def git_status(self, parent):
    """Call git status on the current directory."""
    l = Gio.SubprocessLauncher()
    l.set_flags(Gio.SubprocessFlags.STDOUT_PIPE | Gio.SubprocessFlags.STDERR_PIPE)
    l.set_cwd(self.directory.get_path())
    p = l.spawnv(['git', 'status', '--porcelain', '.'])
    p.communicate_async(None, None, self.git_status_callback, parent)

  def on_menu_activate(self, w, m, key, gfile):
    """Callback for menu item being activated."""
    fname = f'on_{key}_activate'
    f = getattr(self, fname)
    f(m, gfile)

  def on_toolbar_clicked(self, w, b, key):
    """Callback for toolbar button being clicked."""
    fname = f'on_{key}_clicked'
    f = getattr(self, fname)
    f(b)

  def on_up_clicked(self, b):
    parent = self.directory.get_parent()
    if parent:
      self.browse(parent)

  def on_terminal_clicked(self, b):
    self.emit('terminal-activated', self.directory)

  def on_hidden_clicked(self, b):
    self.show_hidden = b.get_active()
    self.refresh()

  def on_refresh_clicked(self, b):
    self.refresh()

  def on_row_activated(self, w, path, column):
    giter = self.model.get_iter(path)
    f = self.model.get_value(giter, 0)
    if f.is_directory:
      self.emit('directory-activated', f.file)
    else:
      child = self.directory.get_child(f.name)
      self.emit('file-activated', child)

  def on_button_press_event(self, w, event):
    if event.button != Gdk.BUTTON_SECONDARY:
      return
    spec = w.get_path_at_pos(int(event.x), int(event.y))
    if not spec:
      return
    path, c, rx, ry = spec
    giter = self.model.get_iter(path)
    f = self.model.get_value(giter, 0)
    if f.is_directory:
      menu = ui.DirectoryPopupMenu(f.file)
    else:
      menu = ui.FilePopupMenu(f.file, suppress=['browseparent'])
    menu.connect('activate', self._on_menu_activate)
    menu.popup(event)
    return True

  def on_directory_notify(self, w, prop):
    self.title.set_label(self.directory.get_path())
    self.emit('directory-changed', self.directory)


if __name__ == '__main__':
  w = Gtk.Window()
  w.connect('destroy', Gtk.main_quit)
  fs = Files(None)
  fs.create_ui()
  w.add(fs)
  w.resize(400, 400)
  w.show_all()

  fs.browse_path('/home/aa/src/bominade')

  def term_activated(w, f):
    print('term-activated', w, f, f.get_path())

  def file_activated(w, f):
    print('file-activated', w, f, f.get_path())

  def directory_changed(w, f):
    print('directory-changed', w, f, f.get_path())

  def directory_activated(w, f):
    print('directory activate')
    w.browse(f)

  def file_destroyed(w, f):
    print('file-destroyed', w, f, f.get_path())

  fs.connect('terminal-activated', term_activated)
  fs.connect('file-activated', file_activated)
  fs.connect('directory-changed', directory_changed)
  fs.connect('directory-activated', directory_activated)
  fs.connect('file-destroyed', file_destroyed)


  Gtk.main()



# (c) 2005-2020 Ali Afshar <aafshar@gmail.com>.
# MIT License. See LICENSE.
# vim: ft=python sw=2 ts=2 sts=2 tw=80

"""Bominade Terminal Emulator"""

import webbrowser, os, pwd

from gi.repository import Gtk, Vte, GLib, GObject, Gio, Gdk, Pango

from b8 import ui


class CwdDiscovery(GObject.GObject):

  __gtype_name__ = 'b8-terminals-cwddiscovery'

  cwd = GObject.Property(type=Gio.File)
  running = GObject.Property(type=bool, default=False)

  __gsignals__ = {
    'cwd-changed': (GObject.SignalFlags.RUN_FIRST, None, (Gio.File,)),
  }

  def __init__(self, pid):
    GObject.GObject.__init__(self)
    self.pid = pid
    self.connect('notify::cwd', self._on_cwd_changed)

  def start(self):
    self.running = True
    self.refresh()
    GLib.timeout_add(1000, self.refresh)

  def stop(self):
    self.running = False

  def refresh(self):
    try:
     l = Gio.File.new_for_path(self.proclink).query_info('*', 0, None)
    except GLib.Error:
      self.stop()
      return False
      
    new = Gio.File.new_for_path(l.get_symlink_target())
    if not self.cwd or (new.get_path() != self.cwd.get_path()):
      self.cwd = new
    return self.running

  def _on_cwd_changed(self, w, prop):
    self.emit('cwd-changed', self.cwd)

  @property
  def proclink(self):
    return f'/proc/{self.pid}/cwd'


class TerminalTheme(GObject.GObject):

  def __init__(self, name):
    GObject.GObject.__init__(self)
    self.default_theme = TERMINAL_THEMES.get('b8')
    raw_theme = TERMINAL_THEMES.get(name)
    if not raw_theme:
      self.error(f'theme {theme_name} does not exist, using default')
      raw_theme = self.default_theme
    self.parse(raw_theme)
    self.font_name = 'Monospace 13'
    self.font_desc = Pango.font_description_from_string(self.font_name)

  def parse(self, theme):
    self.foreground = Gdk.RGBA()
    self.foreground.parse(theme['foreground'])
    self.background = Gdk.RGBA()
    self.background.parse(theme['background'])
    self.cursor = Gdk.RGBA()
    self.cursor.parse(theme.get('cursor', theme['foreground']))
    self.palette = []
    theme_palette = theme['palette']
    for s in theme_palette:
      c = Gdk.RGBA()
      c.parse(s)
      self.palette.append(c)


class Terminals(Gtk.Notebook, ui.MenuHandlerMixin):

  __gtype_name__ = 'b8-terminals'

  __gsignals__ = {
    'file-activated': (GObject.SignalFlags.RUN_FIRST, None, (Gio.File,)),
    'file-destroyed': (GObject.SignalFlags.RUN_FIRST, None, (Gio.File,)),
    'terminal-activated': (GObject.SignalFlags.RUN_FIRST, None, (Gio.File,)),
    'directory-activated': (GObject.SignalFlags.RUN_FIRST, None, (Gio.File,)),
  }

  def __init__(self, font, theme):
    Gtk.Notebook.__init__(self)
    self.set_tab_pos(Gtk.PositionType.BOTTOM)
    self.set_scrollable(True)
    self.font = Pango.font_description_from_string(font)
    self.theme = TerminalTheme(theme)

  def create(self, wd=None):
    if not wd:
      wd = os.path.expanduser('~')
    t = Terminal()
    t.term.set_font(self.font)
    t.apply_theme(self.theme)
    t.start(wd)
    self.append(t)

  def append(self, t):
    pagenum = self.append_page(t, t.label)
    self.show_all()
    self.set_current_page(pagenum)

  def prev(self):
    c = self.get_current_page()
    n = c - 1
    if n < 0:
      n = self.get_n_pages() - 1
    self.change_tab(n)
    
  def next(self):
    c = self.get_current_page()
    n = c + 1
    if n == self.get_n_pages():
      n = 0
    self.change_tab(n)

  def change_tab(self, n):
    self.set_current_page(n)
    p = self.get_nth_page(n)
    if hasattr(p, 'term'):
      p = p.term
    p.grab_focus()

  def remove_terminal(self, t):
    if not self.get_n_pages():
      self.create()

  def shutdown(self):
    print('shutdown', self.get_children())
    for child in self.get_children():
      child.term.disconnect(child.exited_handler)
      child.cwd_discovery.stop()
      if child.pid > 0:
        os.kill(child.pid, 15)



    


class Terminal(Gtk.HBox):

  __gtype_name__ = 'b8-terminals-terminal'


  file_match = GObject.Property(type=int)
  url_match = GObject.Property(type=int)
  cwd_discovery = GObject.Property(type=CwdDiscovery)
  label = GObject.Property(type=Gtk.Label)
  pid = GObject.Property(type=int)
  cwd = GObject.Property(type=Gio.File)
  exited_handler = GObject.Property(type=int)

  def __init__(self):
    Gtk.HBox.__init__(self)
    self.term = Vte.Terminal()
    sw = Gtk.ScrolledWindow()
    sw.add(self.term)
    self.pack_start(sw, True, True, 0)
    self.pack_start(self._create_toolbar(), False, False, 0)
    self._add_matches()
    self.label = self._create_tab_label()
    self.term.connect('button-press-event', self._on_button_press_event)
    self.exited_handler = self.term.connect('child-exited', self._on_child_exited)

  def _create_tab_label(self):
    self.label = Gtk.Label()
    self.label.set_ellipsize(Pango.EllipsizeMode.START)
    self.label.set_width_chars(18)
    return self.label

  def _update_label(self):
    self.label.set_markup(self._markup())

  def apply_theme(self, theme):
    self.term.set_colors(theme.foreground, theme.background, theme.palette)
    self.term.set_color_cursor(theme.cursor)
    self.term.set_color_cursor_foreground(theme.cursor)
    self.term.set_font(theme.font_desc)

  def _markup(self):
    ecwd = GLib.markup_escape_text(self.cwd.get_path())
    return f'<span size="small">{ecwd} [<span weight="bold">{self.pid}</span>]</span>'
    
  def _create_toolbar(self):
    t = ui.MiniToolbar.vertical(
        [
          ui.ImageButton(
            key='terminal',
            icon='utilities-terminal',
            tooltip='New terminal in the same directory',
          ),
          ui.ImageButton(
            key='browse',
            icon='folder-open',
            tooltip='Browse the current working directory',
          ),
          ui.ImageButton(
            key='copy',
            icon='edit-copy',
            tooltip='Copy selected text to clipboard',
          ),
          ui.ImageButton(
            key='paste',
            icon='edit-paste',
            tooltip='Paste the clipboard at the cursor',
          ),
          ui.ImageButton(
            key='selectall',
            icon='edit-select-all',
            tooltip='Select all the text in the terminal',
          ),
          Gtk.Frame(),
        ]
    )
    t.connect('clicked', self._on_toolbar_clicked)
    return t


  def _on_toolbar_clicked(self, w, b, key):
    print(b, key)
    handlers = {
        'browse': self._on_browse_clicked,
        'terminal': self._on_terminal_clicked,
        'copy': self._on_copy_clicked,
        'paste': self._on_paste_clicked,
        'selectall': self._on_selectall_clicked,
    }
    f = handlers.get(key)
    if f:
      f(b)

  def _on_browse_clicked(self, w):
    self.get_parent().emit('directory-activated', self.cwd)

  def _on_copy_clicked(self, w):
    self.term.copy_clipboard_format(Vte.Format.TEXT)

  def _on_paste_clicked(self, w):
    self.term.paste_clipboard()

  def _on_selectall_clicked(self, w):
    self.term.select_all()

  def _on_terminal_clicked(self, w):
    self.get_parent().emit('terminal-activated', self.cwd)

  def _add_matches(self):
    self.url_match = self.term.match_add_regex(URL_RE, 0)
    self.file_match = self.term.match_add_regex(FILE_RE, 0)
    for m in [self.url_match, self.file_match]:
      self.term.match_set_cursor_name(m, MATCH_CURSOR)

  def _started_callback(self, t, pid, *args):
    self.pid = pid
    self.cwd_discovery = CwdDiscovery(pid)
    self.cwd_discovery.connect('cwd-changed', self._on_cwd_changed)
    self.cwd_discovery.start()
    self.term.watch_child(pid)
    #self._update_label()
    #self.update_label()
    #GLib.timeout_add(1000, self.label_updater)
    self.grab_focus()

  def _on_cwd_changed(self, w, cwd):
    self.cwd = cwd
    self._update_label()

  def start(self, wd):
    self.term.spawn_async(
        Vte.PtyFlags.DEFAULT,
        wd,
        [self._get_default_shell()],
        [],
        GLib.SpawnFlags.DEFAULT | GLib.SpawnFlags.DO_NOT_REAP_CHILD,
        None,
        None,
        -1,
        None,
        self._started_callback,
        None)
    self.started = True
    self.term.grab_focus()

  def _get_default_shell(self):
    """Returns the default shell for the user"""
    # Environ, or fallback to login shell
    return os.environ.get('SHELL', pwd.getpwuid(os.getuid())[-1])

  def _get_selection(self):
    if self.term.get_has_selection():
      # Get the selection value from the primary buffer
      clipboard = Gtk.Clipboard.get(Gdk.SELECTION_PRIMARY)
      selection = clipboard.wait_for_text()
      return selection

  def _on_button_press_event(self, w, event):
    # First check the selection
    m, tag = self._get_selection(), self.file_match
    if not m:
      m, tag = w.match_check_event(event)
    if not m:
      return
    m = m.strip()
    tag_handlers = {
        self.file_match: self._on_match_fs_click,
        self.url_match: self._on_match_url_click,
    }
    f = tag_handlers.get(tag)
    if f:
      f(m, event)

  def _on_match_fs_click(self, m, event):
    m = os.path.expanduser(m)
    if m.startswith('/'):
      f = Gio.File.new_for_path(m)
    else:
      f = self.cwd.get_child(m)
    if not f.query_exists(None):
      return
    if f.query_file_type(0, None) == Gio.FileType.DIRECTORY:
      if event.button == Gdk.BUTTON_PRIMARY:
        self.get_parent().emit('directory-activated', f)
      elif event.button == Gdk.BUTTON_SECONDARY:
        menu = ui.DirectoryPopupMenu(f)
        menu.connect('activate', self.get_parent()._on_menu_activate)
        menu.popup(event)
    else:
      if event.button == Gdk.BUTTON_PRIMARY:
        self.get_parent().emit('file-activated', f)
      elif event.button == Gdk.BUTTON_SECONDARY:
        menu = ui.FilePopupMenu(f)
        menu.connect('activate', self.get_parent()._on_menu_activate)
        menu.popup(event)

  def _on_match_url_click(self, m, event):
    webbrowser.open(m)

  def _on_child_exited(self, t, status):
    self.pid = -1
    t.feed(f'\x1b[0;1;34mExited, '
                   f'status: \x1b[0;1;31m{status} \r\n'
                   f'\x1b[0mpress enter to close'.encode('utf-8'))
    t.connect('key-press-event', self._on_keypress_after_exit)

  def _on_keypress_after_exit(self, t, event):
    key_name = Gdk.keyval_name(event.keyval)
    if key_name == 'Return':
      self.close()

  def close(self):
    self.get_parent().remove_terminal(self)


# https://github.com/luvit/pcre2/blob/master/src/pcre2.h.in
PCRE2_MULTILINE = 0x00000400

FILE_RE = Vte.Regex.new_for_match(
  r'(\H+)',
  -1,
  PCRE2_MULTILINE,
)

URL_RE = Vte.Regex.new_for_match(
  (r'https?:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\.'
   r'[a-zA-Z0-9()]{1,6}\b([-a-zA-Z0-9()@:%_\+.~#?&//=]*)'),
  -1,
  PCRE2_MULTILINE,
)

MATCH_CURSOR = 'pointer'


TERMINAL_THEMES = {
  "b8": {
    "name": "b8",
    "foreground": "#839496",
    "background": "#2e2e2e",
    "cursor": "#f0544c",
    "activity": "#dc322f",
    "palette": [
      "#073642",
      "#dc322f",
      "#859900",
      "#b58900",
      "#268bd2",
      "#d33682",
      "#2aa198",
      "#eee8d5",
      "#002b36",
      "#cb4b16",
      "#586e75",
      "#657b83",
      "#839496",
      "#6c71c4",
      "#93a1a1",
      "#fdf6e3"
    ]
  },
  "xubuntu_dark": {
    "name": "xubuntu_dark",
    "foreground": "#b7b7b7",
    "background": "#131926",
    "cursor": "#0f4999",
    "activity": "#0f4999",
    "palette": [
      "#000000",
      "#aa0000",
      "#44aa44",
      "#aa5500",
      "#0039aa",
      "#aa22aa",
      "#1a92aa",
      "#aaaaaa",
      "#777777",
      "#ff8787",
      "#4ce64c",
      "#ded82c",
      "#295fcc",
      "#cc58cc",
      "#4ccce6",
      "#ffffff"
    ]
  },
  "solarized_dark": {
    "name": "solarized_dark",
    "foreground": "#839496",
    "background": "#002b36",
    "cursor": "#93a1a1",
    "activity": "#dc322f",
    "palette": [
      "#073642",
      "#dc322f",
      "#859900",
      "#b58900",
      "#268bd2",
      "#d33682",
      "#2aa198",
      "#eee8d5",
      "#002b36",
      "#cb4b16",
      "#586e75",
      "#657b83",
      "#839496",
      "#6c71c4",
      "#93a1a1",
      "#fdf6e3"
    ]
  },
  "white_on_black": {
    "name": "white_on_black",
    "foreground": "#ffffff",
    "background": "#000000",
    "palette": []
  },
  "black_on_white": {
    "name": "black_on_white",
    "foreground": "#000000",
    "background": "#ffffff",
    "palette": []
  },
  "green_on_black": {
    "name": "green_on_black",
    "foreground": "#17f018",
    "background": "#000000",
    "palette": []
  },
  "xubuntu_light": {
    "name": "xubuntu_light",
    "foreground": "#1f3566",
    "background": "#f1f1f1",
    "cursor": "#0f4999",
    "activity": "#0f4999",
    "palette": [
      "#000000",
      "#aa0000",
      "#44aa44",
      "#aa5500",
      "#0039aa",
      "#aa22aa",
      "#1a92aa",
      "#aaaaaa",
      "#888888",
      "#ff8787",
      "#4ce64c",
      "#ded82c",
      "#295fcc",
      "#cc58cc",
      "#4ccce6",
      "#ffffff"
    ]
  },
  "tango": {
    "name": "tango",
    "foreground": "#ffffff",
    "background": "#000000",
    "palette": [
      "#000000",
      "#cc0000",
      "#4e9a06",
      "#c4a000",
      "#3465a4",
      "#75507b",
      "#06989a",
      "#d3d7cf",
      "#555753",
      "#ef2929",
      "#8ae234",
      "#fce94f",
      "#739fcf",
      "#ad7fa8",
      "#34e2e2",
      "#eeeeec"
    ]
  },
  "solarized_light": {
    "name": "solarized_light",
    "foreground": "#073642",
    "background": "#fdf6e3",
    "cursor": "#073642",
    "activity": "#dc322f",
    "palette": [
      "#073642",
      "#dc322f",
      "#859900",
      "#b58900",
      "#268bd2",
      "#d33682",
      "#2aa198",
      "#eee8d5",
      "#002b36",
      "#cb4b16",
      "#586e75",
      "#657b83",
      "#839496",
      "#6c71c4",
      "#93a1a1",
      "#fdf6e3"
    ]
  },
  "dark_pastels": {
    "name": "dark_pastels",
    "foreground": "#dcdcdc",
    "background": "#2c2c2c",
    "cursor": "#dcdcdc",
    "palette": [
      "#3f3f3f",
      "#705050",
      "#60b48a",
      "#dfaf8f",
      "#9ab8d7",
      "#dc8cc3",
      "#8cd0d3",
      "#dcdcdc",
      "#709080",
      "#dca3a3",
      "#72d5a3",
      "#f0dfaf",
      "#94bff3",
      "#ec93d3",
      "#93e0e3",
      "#ffffff"
    ]
  },
  "xterm": {
    "name": "xterm",
    "foreground": "#000000",
    "background": "#ffffff",
    "palette": [
      "#000000",
      "#cd0000",
      "#00cd00",
      "#cdcd00",
      "#0000cd",
      "#cd00cd",
      "#00cdcd",
      "#e5e5e5",
      "#7f7f7f",
      "#ff0000",
      "#00ff00",
      "#ffff00",
      "#5c5cff",
      "#ff00ff",
      "#00ffff",
      "#ffffff"
    ]
  }
}

if __name__ == '__main__':
  w = Gtk.Window()
  w.connect('destroy', Gtk.main_quit)
  w.resize(400, 400)
  t = Terminals()
  t.create('/home/aa/src/a8')
  w.add(t)

  fs = t

  w.show_all()
  def term_activated(w, f):
    print('term-activated', w, f, f.get_path())

  def file_activated(w, f):
    print('file-activated', w, f, f.get_path())

  def directory_activated(w, f):
    print('directory activate', f.get_path())

  def file_destroyed(w, f):
    print('file-destroyed', w, f, f.get_path())

  fs.connect('terminal-activated', term_activated)
  fs.connect('file-activated', file_activated)
  fs.connect('directory-activated', directory_activated)
  fs.connect('file-destroyed', file_destroyed)
  Gtk.main()



# (c) 2005-2020 Ali Afshar <aafshar@gmail.com>.
# MIT License. See LICENSE.
# vim: ft=python sw=2 ts=2 sts=2 tw=80

import gi

gi.require_version("Gtk", "3.0")
gi.require_version("PangoCairo", "1.0")
gi.require_version("Vte", "2.91")

import os, sys
from gi.repository import GObject, GLib, Gio, Gtk, Gdk

from b8 import logs, configs, vim, files, buffers, terminals


class B8Window(Gtk.Window, logs.LoggerMixin):

  __gtype_name__ = 'b8-app-window'

  def __init__(self, b8):
    Gtk.Window.__init__(self)
    logs.LoggerMixin.__init__(self)
    self.b8 = b8
    self.set_icon_name('media-seek-forward')
    self.set_title('b8 ♡ u')
    hsplit = Gtk.HPaned()
    lsplit = Gtk.VPaned()
    rsplit = Gtk.VPaned()
    self.rsplit = rsplit

    lsplit.pack1(self.b8.buffers, resize=True, shrink=False)
    lsplit.pack2(self.b8.files, resize=True, shrink=False)
    hsplit.pack1(lsplit, resize=True, shrink=False)
    hsplit.pack2(rsplit, resize=True, shrink=False)
    hsplit.set_property('position', 250)
    self.add(hsplit)

    rsplit.set_property('position', 600)
    rsplit.pack1(self.b8.vim, resize=True, shrink=False)
    rsplit.pack2(self.b8.terminals, resize=True, shrink=False)
    self.resize(1024, 900)

    self.connect('delete-event', self.b8._on_delete_event)


class B8(GObject.GObject, logs.LoggerMixin):

  __gtype_name__ = 'b8-app'

  def __init__(self):
    GObject.GObject.__init__(self)
    logs.LoggerMixin.__init__(self)
    self.config = configs.Config()
    self._add_actions()
    self.vim = vim.Embedded()
    self.vim.connect('ready', self._on_vim_ready)
    self.vim.connect('exited', self._on_vim_exited)
    self.buffers = buffers.Buffers()
    self.files = files.Files()
    self.terminals = terminals.Terminals(
        font=self.config.get(('terminal', 'font')),
        theme=self.config.get(('terminal', 'theme')),
    )

    for w in [self.buffers, self.files, self.terminals]:
      w.connect('directory-activated', self._on_directory_activated)
      w.connect('file-activated', self._on_file_activated)
      w.connect('terminal-activated', self._on_terminal_activated)
      w.connect('file-destroyed', self._on_file_destroyed)

    self.vim.connect('buffer-changed', self._on_buffer_changed)
    self.vim.connect('buffer-deleted', self._on_buffer_deleted)
    self.buffers.connect('buffer-activated', self._on_buffer_activated)

  def _on_directory_activated(self, w, f):
    self.files.browse(f)

  def _on_file_activated(self, w, f):
    self.vim.open_buffer(f.get_path())

  def _on_buffer_changed(self, w, bnum, f):
    self.buffers.change(f, bnum)
    self.window.set_title(f.get_path())

  def _on_buffer_deleted(self, w, bnum, f):
    self.buffers.remove(f, bnum)

  def _on_file_destroyed(self, w, f):
    self.vim.close_buffer(f.get_path())

  def _on_buffer_activated(self, w, b):
    self.vim.change_buffer(b.number)

  def _on_terminal_activated(self, w, f):
    self.terminals.create(f.get_path())

  def _add_actions(self):
    self.keymap = {}
    config_map = {
        'next-buffer': self._on_nextbuffer_activate,
        'previous-buffer': self._on_prevbuffer_activate,
        'next-terminal': self._on_nextterminal_activate,
        'previous-terminal': self._on_prevterminal_activate,
        'new-terminal': self._on_newterminal_activate,
    }
    for act in config_map:
      accel = self.config.get(('shortcuts', act))
      km = Gtk.accelerator_parse(accel)
      self.keymap[km] = config_map[act]

  def _on_newterminal_activate(self):
    self.terminals.create()

  def _on_prevterminal_activate(self):
    self.terminals.prev()

  def _on_nextterminal_activate(self):
    self.terminals.next()

  def _on_nextbuffer_activate(self):
    self.buffers.next()

  def _on_prevbuffer_activate(self):
    self.buffers.prev()

  def _on_key_press_event(self, w, event):
    kn = Gdk.keyval_name(event.keyval)
    if kn in vim.MODIFIER_NAMES:
      return
    state = event.state & Gdk.ModifierType.MOD1_MASK
    act = self.keymap.get(
        (event.keyval,
         event.state & Gdk.ModifierType.MOD1_MASK))
    if act:
      act()
      return True

  def _on_vim_ready(self, w):
    self.debug('vim is ready')
    self.window.show_all()
    self.vim.grab_focus()
    self.window.present()

  def _on_vim_exited(self, w):
    self.debug('goodbye, b8 ♡ u')
    Gtk.main_quit()

  def _on_delete_event(self, wi, event):
    self.quit()
    return True

  def run(self):
    self.debug('activating')
    self.window = B8Window(self)
    self.window.connect('key-press-event', self._on_key_press_event)
    self._add_actions()
    self.vim.start()
    self.terminals.create(os.path.expanduser('~'))
    self.files.browse_path(os.path.expanduser('~'))
    Gtk.main()

  def quit(self):
    self.terminals.shutdown()
    self.vim.quit()



def main():
  app = B8()
  app.run()


if __name__ == '__main__':
  main()




# (c) 2005-2020 Ali Afshar <aafshar@gmail.com>.
# MIT License. See LICENSE.
# vim: ft=python sw=2 ts=2 sts=2 tw=80

"""Embeddable NeoVim for PyGTK.

This module can be used entirely on its own separate from B8. This is a basic
requirement of the thing. It is also a clean-room implementation as it uses
NeoVim's `ext_linegrid` as the older methods are depracated and the existing
PyGTK widget for NeoVim uses the depracated method.

Consider the bare minimum script:

    w = Gtk.Window()
    w.connect('destroy', Gtk.main_quit)
    w.resize(800, 600)

    vim = Embedded()
    w.add(vim)


    def on_vim_started(widget):
      # Show the window once Vim has started and given use the right Gui options,
      # i.e. the font.
      w.show_all()

    # the `started` signal tells us Vim is up and running and configured enough to
    # display.
    vim.connect('started', on_vim_started)

    Gtk.main()

The startup flow is a bit annoying because we don't want to display the widget
until Vim is started and has told us the value of `guifont` otherwise we will
guess a size of the Vim grid and then have tor esize in front of the user which
will be janky at least. In reality it's worse than just jank. We don't have
enough data to render a widget so we'd have to just show a blank screen.
"""

from typing import Iterable, List
import msgpack
from gi.repository import Gio, GLib, GObject, Gdk, Gtk, Pango, PangoCairo
import cairo

from b8 import ui, logs


class Grid:
  """NeoVim grid"""

  def __init__(self, width, height):
    self.width = width
    self.height = height
    self.cells = [[Cell() for col in range(self.width)] for col in
        range(self.height)]


class Cell:
  """Single cell within a NeoVim grid"""
  text = ''  
  hl = None

  def copy_to(self, c):
    c.text = self.text
    c.hl = self.hl

  def is_same(self, other):
    return self.text == other.text and self.hl is other.hl


class Mode(GObject.GObject):
  """Information about a NeoVim mode."""

  __gtype_name__ = 'b8-vim-modeinfo'

  name = GObject.Property(type=str)

  mouse_shape = None
  cursor_shape = None
  cell_percentage = None
  blinkwait = None
  blinkon = None
  blinkoff = None
  hl_id = None
  id_lm = None
  attr_id = None
  attr_id_lm = None
  short_name = None

  def __repr__(self):
    return f'<Mode name={self.name} {self.cursor_shape} {self.mouse_shape}>'


class Highlight:
  """Information about a NeoVim highlight."""
  foreground = None 
  background = None
  special = None
  bold = False
  undercurl = False
  reverse = False
  strikethrough = False
  underline = False
  blend = False
  italic = False

  def __repr__(self):
    return f'<Highlight fg={self.foreground} bg={self.background}>'


class Color:
  """NeoVim color with conversion to RGB"""

  def __init__(self, color_value):
    self.r = ((color_value >> 16) & 255) / 256.0
    self.g = ((color_value >> 8) & 255) / 256.0
    self.b = (color_value & 255) / 256.0

  def __repr__(self):
    return f'Color<{self.r} {self.g} {self.b}>'


class Cursor(GObject.GObject):

  __gtype_name__ = 'b8-vim-cursorposition'

  x = GObject.Property(type=int, default=0)
  y = GObject.Property(type=int, default=0)

  def __init__(self, x, y):
    GObject.GObject.__init__(self)
    self.x = x
    self.y = y

  def __repr__(self):
    return f'<Cursor x={self.x} y={self.y}'


class Buffer(GObject.GObject):

  __gtype_name__ = 'b8-vim-buffer'

  number = GObject.Property(type=int)
  file = GObject.Property(type=Gio.File)

  def __init__(self, number, file):
    GObject.GObject.__init__(self)
    self.number = number
    self.file = file
    self.path = file.get_path()
    self.parent = file.get_parent()
    self.name = file.get_basename()
    self.ename = GLib.markup_escape_text(self.name)
    self.eparent = GLib.markup_escape_text(self.parent.get_path())
    self.markup = (f'<span size="medium" weight="bold">{self.ename}</span>\n'
                   f'<span size="x-small">{self.parent.get_path()}</span>')

  @classmethod
  def from_ext_hook(cls, ext_data):
    """Alternative constructor from msgpack protocol."""
    return cls(msgpack.unpackb(ext_data), None)


class Result(GObject.GObject):

  __gtype_name__ = 'b8-vim-result'

  __gsignals__ = {
      'success': (GObject.SignalFlags.RUN_FIRST, None, (object,)),
      'error': (GObject.SignalFlags.RUN_FIRST, None, (object,)),
  }

  def __init__(self, cid, name, args):
    GObject.GObject.__init__(self)
    self.cid = cid
    self.name = name
    self.args = args

  def respond(self, cid, error, success):
    if error:
      self.emit('error', error)
    else:
      self.emit('success', success)


class Drag(GObject.GObject):

  __gtype_name__ = 'b8-vim-drag'

  button = GObject.Property(type=int)
  pressed = GObject.Property(type=bool, default=False)


class Embedded(Gtk.DrawingArea, logs.LoggerMixin):

  __gtype_name__ = 'b8-vim'

  __gsignals__ = {
      'ready': (GObject.SignalFlags.RUN_FIRST, None, ()),
      'exited': (GObject.SignalFlags.RUN_FIRST, None, ()),
      'buffer-changed': (GObject.SignalFlags.RUN_FIRST, None, (int, Gio.File,)),
      'buffer-deleted': (GObject.SignalFlags.RUN_FIRST, None, (int, Gio.File,)),
      'mode-changed': (GObject.SignalFlags.RUN_FIRST, None, (Mode,)),
      'cursor-changed': (GObject.SignalFlags.RUN_FIRST, None, (Cursor,)),
  }

  cursor = GObject.Property(type=Cursor, default=Cursor(0, 0))
  width = GObject.Property(type=int, default=20)
  height = GObject.Property(type=int, default=20)
  cid = GObject.Property(type=int, default=0)
  mode = GObject.Property(type=Mode)
  proc = GObject.Property(type=Gio.Subprocess)
  source = GObject.Property(type=GLib.Source)

  button_drag = False
  button_pressed = False

  def __init__(self):
    Gtk.DrawingArea.__init__(self)
    logs.LoggerMixin.__init__(self)
    self.unpacker = msgpack.Unpacker(ext_hook=self._ext_hook, raw=False)
    self.options = {}
    self.highlights = {}
    self.mode = None
    self.modes = {}
    self.pending_commands = {}
    self.drag = Drag()
    self.default_highlight = None
    self.button_pressed = None
    self.set_can_focus(True)
    self.add_events(Gdk.EventMask.KEY_PRESS_MASK |
                                 Gdk.EventMask.BUTTON_PRESS_MASK |
                                 Gdk.EventMask.BUTTON_RELEASE_MASK |
                                 Gdk.EventMask.POINTER_MOTION_MASK |
                                 Gdk.EventMask.SCROLL_MASK |
                                 Gdk.EventMask.FOCUS_CHANGE_MASK)
    self.connect('size-allocate', self._on_size_allocate)
    self.connect('realize', self._on_realize)
    self.connect('draw', self._on_draw)
    self.connect('key-press-event', self._on_key_press_event)
    self.connect('button-press-event', self._on_button_press_event)
    self.connect('button-release-event', self._on_button_release_event)
    self.connect('motion-notify-event', self._on_motion_notify_event)
    self.connect('focus-in-event', self._on_focus_in_event)
    self.connect('focus-out-event', self._on_focus_out_event)
    self.connect('scroll-event', self._on_scroll_event)
    self.connect('notify::mode', self._on_notify_mode)
    self.connect('notify::cursor', self._on_notify_cursor)

  def quit(self):
    self.debug('quitting')
    self._cmd('nvim_command', ['q!']);

  def open_buffer(self, path):
    self._cmd('nvim_command', [f'e!{path}']);

  def change_buffer(self, bnum):
    self._cmd('nvim_command', [f'b!{bnum}']);

  def close_buffer(self, path):
    self._cmd('nvim_command', [f'confirm bd{path}']);

  def command(self, name, *args):
    return self._cmd(name, list(args))

  def start(self):
    self._start()

  def _ext_hook(self, ext_type, ext_data):
    """Called when NeoVim sends extended type information."""
    if ext_type == 0:
      return Buffer.from_ext_hook(ext_data)

  def _generate_message(self, name: str, args: list):
    self.cid = (self.cid + 1) % 256
    return [0, self.cid, name, args]

  def _serialize_message(self, msg: list) -> str:
    return msgpack.dumps(msg)

  def _cmd(self, name: str, args: list) -> int:
    msg = self._generate_message(name, args)
    d = self._serialize_message(msg)
    self.vim_in.write_all(d, None)
    r = self.pending_commands[self.cid] = Result(self.cid, name, args)
    return r

  def _out_callback(self, *args):
    if self.vim_out.is_readable():
      d = self.vim_out.read_bytes(1024)
      self.unpacker.feed(d.get_data())
      for msg in self.unpacker:
        self._msg_callback(msg)
    return True

  def _msg_callback(self, msg):
    msg_handlers = {
        1: self._reply_callback,
        2: self._notification_callback,
    }
    msg_handlers[msg[0]](msg[1:])

  def _reply_callback(self, msg):
    rid, err, result = msg
    result = self.pending_commands.pop(rid)
    result.respond(rid, err, result)
    self.debug(f'reply {msg}')

  def _notification_callback(self, msg):
    msg_handlers = {
        'buffers': self._buffers_callback,
        'system': self._system_callback,
        'redraw': self._redraw_callback,
    }
    msg_handlers[msg[0]](msg[1])

  def _buffers_callback(self, msg):
    action, bs, path = msg
    if not path:
      return
    bnum = int(bs)
    gf = Gio.File.new_for_path(path)
    msg_handlers = {
        'enter': self._buffers_enter_callback,
        'delete': self._buffers_delete_callback,
    }
    f = msg_handlers.get(action)
    if f:
      f(bnum, gf)
    else:
      print('unhandled buffers', msg)
    print('buffers', msg)

  def _buffers_enter_callback(self, bnum, f):
    self.emit('buffer-changed', bnum, f)
  
  def _buffers_delete_callback(self, bnum, f):
    self.emit('buffer-deleted', bnum, f)
  
  def _system_callback(self, msg):
    action = msg[0]
    msg_handlers = {
        'leave': self._system_leave_callback,
        'enter': self._system_enter_callback,
    }
    f = msg_handlers.get(action)
    if f:
      f()
    else:
      print('unhandled system', msg)

  def _system_leave_callback(self):
    """Called for a Vim VimLeave notification."""
    self.emit('exited')

  def _system_enter_callback(self):
    self.debug('VimEnter autocmd')
    if not self.options.get('guifont'):
      self.debug('vim has not set guifont, so we will')
      self.options['guifont'] = 'Monospace 13'
    self.debug('guifont', data=self.options['guifont'])
    self._calculate_font_size()
    self.emit('ready')

  def _redraw_callback(self, msgs):
    """Called for a Vim redraw notification."""
    msg_handlers = {
        'grid_resize': self._grid_resize_callback,
        'option_set': self._option_set_callback,
        'default_colors_set': self._default_colors_set_callback,
        'hl_attr_define': self._hl_attr_define_callback,
        'grid_cursor_goto': self._grid_cursor_goto_callback,
        'mode_info_set': self._mode_info_set_callback,
        'mode_change': self._mode_change_callback,
        'grid_line': self._grid_line_callback,
        'grid_scroll': self._grid_scoll_callback,
        'flush': self._flush_callback,
    }
    for msg in msgs:
      f = msg_handlers.get(msg[0])
      if f:
        f(*msg[1:])
      else:
        self.debug(f'redraw unhandled {msg[0]}, {len(msg)}')

  def _grid_resize_callback(self, msg):
    gid, cols, rows = msg
    self.grid = Grid(cols, rows)

  def _option_set_callback(self, *args):
    self.options.update(args)
    self.debug('option_set', data=args)

  def _default_colors_set_callback(self, hl):
    fg, bg, special, tfg, tbg = hl
    c = self.default_highlight = Highlight()
    c.foreground = Color(fg)
    c.background = Color(bg)
    c.special = Color(special)

  def _hl_attr_define_callback(self, *args):
    for hl_id, cs, tcs, empty in args:
      c = self.highlights[hl_id] = Highlight()
      for k in cs:
        v = cs[k]
        if isinstance(v, int):
          v = Color(v)
        setattr(c, k, v)

  def _mode_info_set_callback(self, *args):
    modes = args[0][1]
    for mode in modes:
      m = Mode()
      for k in mode:
        v = mode[k]
        setattr(m, k, v)
      self.modes[m.name] = m

  def _mode_change_callback(self, msg):
    mode_name = msg[0]
    mode_id = msg[1]
    self.mode = self.modes[mode_name]

  def _grid_cursor_goto_callback(self, msg):
    gid, rows, cols = msg
    self.cursor = Cursor(cols, rows)

  def _grid_line_callback(self, *args):
    for arg in args:
      row = arg[1]
      colstart = arg[2]
      cells = arg[3]
      last_hl = -1
      for cell in cells:
        text = cell[0]
        hl = last_hl
        repeat = 1
        if len(cell) > 1:
          hl = cell[1]
        if len(cell) > 2:
          repeat = cell[2]
        last_hl = hl
        for i in range(repeat):
          c = self.grid.cells[row][colstart]
          c.text = text
          c.hl = self.highlights.get(hl)
          colstart += 1

  def _grid_scoll_callback(self, msg):
    gid, top, bottom, left, right, rows, cols = msg
    if rows > 0:
      row = top
      while row <= bottom - rows:
        col = left
        while col < right:
          c = self.grid.cells[row][col]
          nc = self.grid.cells[min(row+rows, self.height-1)][col]
          c.text = nc.text
          c.hl = nc.hl
          col += 1
        row += 1
    else:
      row = bottom - 1
      while row >= top - rows:
        col = left
        while col < right:
          c = self.grid.cells[row][col]
          nc = self.grid.cells[row+rows][col]
          c.text = nc.text
          c.hl = nc.hl
          col += 1
        row -= 1

  def _flush_callback(self, *args):
    self.queue_draw()

  def _start(self):
    self.proc = Gio.Subprocess.new(['nvim', '--embed'],
        Gio.SubprocessFlags.STDOUT_PIPE | Gio.SubprocessFlags.STDERR_PIPE |
        Gio.SubprocessFlags.STDIN_PIPE)
    self.vim_in = self.proc.get_stdin_pipe()
    self.vim_out = self.proc.get_stdout_pipe()
    self.source = self.vim_out.create_source()
    self.source.set_callback(self._out_callback)
    self.source.attach(None)

    self._vim_attach()
    self._vim_subscribe()

  def _calculate_font_size(self):
    self.font_name = self.options['guifont']
    self.font_desc = Pango.font_description_from_string(self.font_name)
    sfc = cairo.ImageSurface(cairo.FORMAT_RGB24, 300, 300)
    cr = cairo.Context(sfc)
    layout = PangoCairo.create_layout(cr)
    layout.set_font_description(self.font_desc)
    layout.set_alignment(Pango.Alignment.LEFT)
    layout.set_markup('<span>M</span>')
    self.cell_width, self.cell_height = layout.get_pixel_size()

  def _on_size_allocate(self, w, alloc):
    self.width = int(alloc.width / self.cell_width)
    self.height = int(alloc.height / self.cell_height)
    self._vim_resize()

  def _on_realize(self, w):
    self.pango_layout = PangoCairo.create_layout(w.get_window().cairo_create())
    self.pango_layout.set_alignment(Pango.Alignment.LEFT)
    self.pango_layout.set_font_description(self.font_desc)

  def _on_key_press_event(self, widget, event, *args):
    key_name = Gdk.keyval_name(event.keyval)
    # Fail fast on a known modifier
    if key_name in MODIFIER_NAMES:
      return
    utf8 = chr(Gdk.keyval_to_unicode(event.keyval))
    # Default to the character
    input_str = utf8
    # Known named keys
    if key_name in KEY_NAMES:
      input_str = KEY_NAMES[key_name]
    # Convert to <> format
    needs_stringify = (
      (key_name in KEY_NAMES) |
      event.state & Gdk.ModifierType.CONTROL_MASK |
      event.state & Gdk.ModifierType.MOD1_MASK
    )
    if needs_stringify:
      input_str = self._key_input(input_str, event.state)


    self.debug(f'keypress chr:{input_str} '
               f'utf8:{repr(utf8)} '
               f'name:{key_name} '
               f'state:{event.state}')

    if input_str == '\x00':
      self.error(f'empty string {key_name}')
    self._vim_input(input_str)
    return True


  def _on_button_press_event(self, widget, event, *args):
    self.grab_focus()
    if event.type != Gdk.EventType.BUTTON_PRESS:
      return
    
    button = BUTTON_NAMES.get(event.button)
    if not button:
      return
    
    mod, row, col = self._parse_mouse(event)
    self._vim_input_mouse(button, 'press', mod, row, col)
    self.button_pressed = button
    return True

  def _on_button_release_event(self, widget, event, *args):
    self.button_pressed = None
    self.button_drag = False

  def _on_motion_notify_event(self, widget, event, *args):
    if not self.button_pressed:
        return
    if not self.button_drag:
      self.button_drag = True
    mod, row, col = self._parse_mouse(event)
    self._vim_input_mouse(self.button_pressed, 'drag', mod, row, col)

  def _on_scroll_event(self, widget, event, *args):
    if event.direction == Gdk.ScrollDirection.UP:
      direction = 'up'
    elif event.direction == Gdk.ScrollDirection.DOWN:
      direction = 'down'
    else:
      return
    mod, row, col = self._parse_mouse(event)
    self._vim_input_mouse('wheel', direction, mod, row, col)


  def _on_focus_in_event(self, widget, event):
    self.queue_draw()

  def _on_focus_out_event(self, widget, event):
    self.queue_draw()

  def _on_notify_mode(self, w, prop):
    self.emit('mode-changed', self.mode)

  def _on_notify_cursor(self, w, prop):
    self.emit('cursor-changed', self.cursor)

  def _parse_mouse(self, event):
    col = int(event.x / self.cell_width)
    row = int(event.y / self.cell_height)
    mods = []
    if event.state & Gdk.ModifierType.SHIFT_MASK:
      out.append('S')
    if event.state & Gdk.ModifierType.CONTROL_MASK:
      out.append('C')
    if event.state & Gdk.ModifierType.MOD1_MASK:
      out.append('A')
    mod = ''.join(mods)
    return mod, row, col

  def _key_input(self, input_str, state):
    out = []
    if state & Gdk.ModifierType.SHIFT_MASK:
      out.append('S')
    if state & Gdk.ModifierType.CONTROL_MASK:
      out.append('C')
    if state & Gdk.ModifierType.MOD1_MASK:
      out.append('A')
    out.append(input_str)
    return f'<{"-".join(out)}>'

  def _on_draw(self, w, cr):
    w.get_window().freeze_updates()
    bg = self.default_highlight.background
    cr.set_source_rgb(bg.r, bg.g, bg.b)
    cr.paint()
    for (cy, row) in enumerate(self.grid.cells):
      for (cx, cell) in enumerate(row):


        x = cx * self.cell_width
        y = cy * self.cell_height

        bg = None
        if cell.hl is None:
          bg = self.default_highlight.background
        else:
          bg = cell.hl.background
        if bg is None:
          bg = self.default_highlight.background

        fg = None
        if cell.hl is None:
          fg = self.default_highlight.foreground
        else:
          fg = cell.hl.foreground
        if fg is None:
          fg = self.default_highlight.foreground


        if cell.hl and cell.hl.reverse:
          fg, bg = bg, fg

        if bg is not self.default_highlight.background:
          cr.move_to(x, y)
          cr.set_source_rgb(bg.r, bg.g, bg.b)
          cr.rectangle(x, y, self.cell_width, self.cell_height)
          cr.fill()

        if cx == self.cursor.x and cy == self.cursor.y:
          cursorColor = self.default_highlight.foreground
          cr.set_source_rgb(cursorColor.r, cursorColor.g, cursorColor.b)
          cursor_width = self.cell_width
          if self.has_focus():
            if self.mode.cell_percentage:
              cursor_width *= (self.mode.cell_percentage / 100.0)
            else:
              fg = self.default_highlight.background
            cr.rectangle(x, y, cursor_width, self.cell_height)
            cr.fill()
          else:
            cr.set_line_width(1.2)
            cr.rectangle(x, y, cursor_width-1, self.cell_height-1)
            cr.stroke()

        if cell.text != ' ':
          cr.move_to(x, y)
          cr.set_source_rgb(fg.r, fg.g, fg.b)
          self.pango_layout.set_text(cell.text, -1)
          PangoCairo.update_layout(cr, self.pango_layout)
          PangoCairo.show_layout(cr, self.pango_layout)
          _, r = self.pango_layout.get_pixel_extents()

    w.get_window().thaw_updates()


  def _vim_attach(self):
    self._cmd('nvim_ui_attach', [self.width, self.height, {'ext_linegrid':
      True}])

  def _vim_subscribe(self):
    types = set()
    for sig in VIM_SIGNALS:
      self._cmd('nvim_command', [VIM_SIGNAL_TEMPLATE.format(*sig)])
      types.add(sig[1])
    for t in types:
      self._cmd('nvim_subscribe', [t]);


  def _vim_resize(self):
    self._cmd('nvim_ui_try_resize', [self.width, self.height])

  def _vim_input(self, keys):
    self._cmd('nvim_input', [keys])

  def _vim_input_mouse(self, button, action, modifier, row, col):
    self._cmd('nvim_input_mouse', [button, action, modifier, 0, row, col])


VIM_SIGNAL_TEMPLATE = 'autocmd {} * call rpcnotify(0, "{}", "{}", {})'

VIM_SIGNALS = [
    ('BufAdd', 'buffers', 'add', 'expand("<abuf>"), expand("<amatch>")'),
    ('BufEnter', 'buffers', 'enter', 'expand("<abuf>"), expand("<amatch>")'),
    ('BufDelete', 'buffers', 'delete', 'expand("<abuf>"), expand("<amatch>")'),
    ('VimLeave', 'system', 'leave', ''),
    ('VimEnter', 'system', 'enter', ''),
]



MODIFIER_NAMES = {
    'Shift_L',
    'Shift_R',
    'Control_L',
    'Control_R',
    'Alt_R',
    'Alt_L'
}

BUTTON_NAMES = {
    1: 'left',
    2: 'middle',
    3: 'right',
}

KEY_NAMES = {
    'less': 'LT',
    'BackSpace': 'BS',
    'Return': 'CR',
    'Escape': 'Esc',
    'Delete': 'Del',
    'Page_Up': 'PageUp',
    'Page_Down': 'PageDown',
    'Enter': 'CR',
    'ISO_Left_Tab': 'Tab',
    'Right': 'Right',
    'Left': 'Left',
    'Up': 'Up',
    'Down': 'Down',
    'Insert': 'Insert',
    'End': 'End',
    'Home': 'Home',
}

if __name__ == '__main__':
  w = Gtk.Window()
  w.connect('destroy', Gtk.main_quit)
  w.resize(400, 400)

  v = Embedded()
  w.add(v)

  def on_list_bufs(r, bufs):
    print(['onlistbufs', bufs])

  def on_buffer_changed(e, bnum, path):
    print('buffer changed', bnum, path)
    v.command('nvim_list_bufs').connect('success', on_list_bufs)

  def on_buffer_deleted(e, bnum, path):
    print('buffer deleted', bnum, path)

  def on_mode(e, mode):
    print(mode)

  def on_exited(e):
    print('vim exited')
    Gtk.main_quit()

  v.connect('buffer-changed', on_buffer_changed)
  v.connect('buffer-deleted', on_buffer_deleted)
  v.connect('exited', on_exited)
  v.connect('mode-changed', on_mode)

  def on_vim_started(e):
    w.show_all()
    print('ready', w)




  v.connect('ready', on_vim_started)
  v._start()


  #GLib.log_writer_standard_streams(
  #    GLib.LogLevelFlags.LEVEL_INFO)

  #s = GLib.String()
  #s.append('ffs')

  #field = GLib.LogField()
  #field.key = 'msg'
  #field.value = s
  #GLib.log_structured_array(GLib.LogLevelFlags.LEVEL_INFO, [field])
  #v._cmd('blah', [1,2,3])

  Gtk.main()



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['b8']

package_data = \
{'': ['*']}

install_requires = \
['msgpack >= 1.0.0']

entry_points = \
{'console_scripts': ['b8 = b8.app:main']}

setup(name='b8',
      version='0.1.10',
      description='b8, Bominade Ultralight NeoVim-based IDE',
      author='Ali Afshar',
      author_email='aafshar@gmail.com',
      url='https://gitlab.com/afshar-oss/b8',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      entry_points=entry_points,
      python_requires='>=3.5',
     )
